class BynderAssetDataError(Exception):
    """
    Raised when values expected to be present in an asset API representation
    from Bynder are missing.
    """


class BynderAssetFileTooLarge(Exception):
    """
    Raised when an asset file being downloaded from Bynder is found to be
    larger than specified in ``settings.BYNDER_MAX_DOWNLOAD_FILE_SIZE``
    """



import mimetypes
import os

from io import BytesIO

import requests

from asgiref.local import Local
from bynder_sdk import BynderClient
from django.conf import settings
from django.core.files import File
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.template.defaultfilters import filesizeformat
from wagtail.models import Collection
from willow import Image

from .exceptions import BynderAssetFileTooLarge


_DEFAULT_COLLECTION = Local()


def download_file(
    url: str, max_filesize: int, max_filesize_setting_name: str
) -> InMemoryUploadedFile:
    name = os.path.basename(url)

    # Stream file to memory
    file = BytesIO()
    for line in requests.get(url, timeout=20, stream=True):
        file.write(line)
        if file.tell() > max_filesize:
            file.truncate(0)
            raise BynderAssetFileTooLarge(
                f"File '{name}' exceeded the size limit enforced by the {max_filesize_setting_name} setting, which is currently set to {filesizeformat(max_filesize)}."
            )

    size = file.tell()
    file.seek(0)

    content_type, charset = mimetypes.guess_type(name)
    return InMemoryUploadedFile(
        file,
        field_name="file",
        name=name,
        content_type=content_type,
        size=size,
        charset=charset,
    )


def download_document(url: str) -> InMemoryUploadedFile:
    max_filesize_setting_name = "BYNDER_MAX_DOCUMENT_FILE_SIZE"
    max_filesize = getattr(settings, max_filesize_setting_name, 5242880)
    return download_file(url, max_filesize, max_filesize_setting_name)


def download_image(url: str) -> InMemoryUploadedFile:
    max_filesize_setting_name = "BYNDER_MAX_IMAGE_FILE_SIZE"
    max_filesize = getattr(settings, max_filesize_setting_name, 5242880)
    return download_file(url, max_filesize, max_filesize_setting_name)


def get_image_info(file: File) -> tuple[int, int, str, bool]:
    willow_image = Image.open(file)
    width, height = willow_image.get_size()
    return (width, height, willow_image.format_name, willow_image.has_animation())


def get_output_image_format(original_format: str, *, is_animated: bool = False) -> str:
    conversions = {
        "avif": "png",
        "bmp": "png",
        "webp": "png",
    }
    if is_animated:
        # Convert non-animated GIFs to PNG as well
        conversions["gif"] = "png"

    # Allow the user to override the conversions
    custom_conversions = getattr(settings, "WAGTAILIMAGES_FORMAT_CONVERSIONS", {})
    conversions.update(custom_conversions)

    return conversions.get(original_format, original_format)


def filename_from_url(url: str) -> str:
    return os.path.basename(url)


def get_bynder_client() -> BynderClient:
    return BynderClient(
        domain=getattr(settings, "BYNDER_DOMAIN", ""),
        permanent_token=getattr(settings, "BYNDER_API_TOKEN", ""),
    )


def get_default_collection() -> Collection:
    """
    Return a Collection object that should be used as the default for images and
    documents created to represent Bynder assets.

    The result is cached for the current thread / asyncio task.
    """
    try:
        return _DEFAULT_COLLECTION.value
    except AttributeError:
        pass

    name = "Imported from Bynder"
    try:
        # Use existing collection
        collection = Collection.objects.get(name=name)
    except Collection.DoesNotExist:
        # Create a new one
        collection = Collection.get_first_root_node().add_child(name=name)

    # Cache result for the current thread / asyncio task
    _DEFAULT_COLLECTION.value = collection
    return collection



import contextlib

from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.db import IntegrityError
from django.shortcuts import redirect

from wagtail_bynder.models import BynderAssetMixin
from wagtail_bynder.utils import get_bynder_client


if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse


class BynderAssetCopyMixin:
    model = type[BynderAssetMixin]

    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)
        bynder_client = get_bynder_client()
        self.asset_client = bynder_client.asset_bank_client

    def build_object_from_data(self, asset_data: dict[str, Any]) -> BynderAssetMixin:
        obj = self.model(bynder_id=asset_data["id"])
        obj.update_from_asset_data(asset_data)
        return obj

    def create_object(self, asset_id: str) -> BynderAssetMixin:
        data = self.asset_client.media_info(asset_id)
        obj = self.build_object_from_data(data)
        try:
            # If the asset finished saving in a different thread during the download/update process,
            # return the pre-existing object
            return self.model.objects.get(bynder_id=asset_id)
        except self.model.DoesNotExist:
            try:
                # Save the new object, triggering transfer of the file to media storage
                obj.save()
            except IntegrityError as integrity_error:
                # It's likely the asset finished saving in a different thread while the file was
                # being transferred to media storage
                try:
                    # Lookup the existing object
                    pre_existing = self.model.objects.get(bynder_id=asset_id)
                except self.model.DoesNotExist:
                    # The IntegrityError must have been caused by a custom field, so reraise
                    raise integrity_error from None
                else:
                    # If the newly-downloaded file was successfully copied to storage, delete it
                    with contextlib.suppress(ValueError, FileNotFoundError):
                        if obj.file.path != pre_existing.file.path:
                            obj.file.delete()
                return pre_existing
        return obj

    def update_object(self, asset_id: str, obj: BynderAssetMixin) -> BynderAssetMixin:
        data = self.asset_client.media_info(asset_id)
        if not obj.is_up_to_date(data):
            obj.update_from_asset_data(data)
            obj.save()
        return obj


class RedirectToBynderMixin:
    def setup(self, request, *args, **kwargs) -> None:
        if self.pk_url_kwarg != "pk":
            kwargs["pk"] = kwargs[self.pk_url_kwarg]
        super().setup(request, *args, **kwargs)

    def get(self, request: "HttpRequest", *args, **kwargs) -> "HttpResponse":
        self.object = self.get_object()
        if asset_id := self.object.bynder_id:
            return redirect(
                "https://" + settings.BYNDER_DOMAIN + f"/media/?mediaId={asset_id}"
            )
        return super().get(request, *args, **kwargs)

    def post(self, request: "HttpRequest", *args, **kwargs) -> "HttpResponse":
        self.object = self.get_object()
        if asset_id := self.object.bynder_id:
            return redirect(
                "https://" + settings.BYNDER_DOMAIN + f"/media/?mediaId={asset_id}"
            )
        return super().post(request, *args, **kwargs)
